# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from bzrformats import chk_map, groupcompress, inventory, osutils
from bzrformats import errors as bzrformats_errors
from bzrformats.inventory import (
    ROOT_ID,
    CHKInventory,
    DuplicateFileId,
    InvalidEntryName,
    Inventory,
    InventoryDirectory,
    InventoryEntry,
    InventoryFile,
    NoSuchId,
    TreeReference,
    _chk_inventory_bytes_to_entry,
    _chk_inventory_entry_to_bytes,
    chk_inventory_bytes_to_utf8name_key,
)

from ..inventory_delta import InventoryDelta
from . import TestCase, TestCaseWithMemoryTransport


class TestInventoryUpdates(TestCase):
    def test_creation_from_root_id(self):
        # iff a root id is passed to the constructor, a root directory is made
        inv = inventory.Inventory(root_id=b"tree-root")
        self.assertNotEqual(None, inv.root)
        self.assertEqual(b"tree-root", inv.root.file_id)

    def test_add_path_of_root(self):
        # if no root id is given at creation time, there is no root directory
        inv = inventory.Inventory(root_id=None)
        self.assertIs(None, inv.root)
        # add a root entry by adding its path
        ie = inv.add_path("", "directory", b"my-root", revision=b"test-rev")
        self.assertEqual(b"my-root", ie.file_id)
        self.assertEqual(ie, inv.root)

    def test_add_path(self):
        inv = inventory.Inventory(root_id=b"tree_root")
        ie = inv.add_path("hello", "file", b"hello-id")
        self.assertEqual(b"hello-id", ie.file_id)
        self.assertEqual("file", ie.kind)

    def test_copy(self):
        """Make sure copy() works and creates a deep copy."""
        inv = inventory.Inventory(root_id=b"some-tree-root")
        inv.add_path("hello", "file", b"hello-id")
        inv2 = inv.copy()
        inv.rename_id(b"some-tree-root", b"some-new-root")
        self.assertEqual(b"some-tree-root", inv2.root.file_id)
        self.assertEqual("hello", inv2.get_entry(b"hello-id").name)

    def test_copy_empty(self):
        """Make sure an empty inventory can be copied."""
        inv = inventory.Inventory(root_id=None)
        inv2 = inv.copy()
        self.assertIs(None, inv2.root)

    def test_copy_copies_root_revision(self):
        """Make sure the revision of the root gets copied."""
        inv = inventory.Inventory(root_id=b"someroot", root_revision=b"therev")
        inv2 = inv.copy()
        self.assertEqual(b"someroot", inv2.root.file_id)
        self.assertEqual(b"therev", inv2.root.revision)

    def test_create_tree_reference(self):
        inv = inventory.Inventory(b"tree-root-123")
        inv.add(
            TreeReference(
                b"nested-id",
                "nested",
                parent_id=b"tree-root-123",
                revision=b"rev",
                reference_revision=b"rev2",
            )
        )

    def test_error_encoding(self):
        inv = inventory.Inventory(b"tree-root")
        inv.add(InventoryFile(b"a-id", "\u1234", b"tree-root"))
        from bzrformats.errors import AlreadyVersionedError

        e = self.assertRaises(
            AlreadyVersionedError,
            inv.add,
            InventoryFile(b"b-id", "\u1234", b"tree-root"),
        )
        self.assertContainsRe(str(e), "\\u1234")

    def test_add_recursive(self):
        parent = InventoryDirectory(b"src-id", "src", b"tree-root")
        child = InventoryFile(b"hello-id", "hello.c", b"src-id")
        inv = inventory.Inventory(b"tree-root")
        inv.add(parent)
        inv.add(child)
        self.assertEqual("src/hello.c", inv.id2path(b"hello-id"))


class TestInventoryEntry(TestCase):
    def test_file_invalid_entry_name(self):
        self.assertRaises(
            InvalidEntryName, inventory.InventoryFile, b"123", "a/hello.c", ROOT_ID
        )

    def test_file_backslash(self):
        file = inventory.InventoryFile(b"123", "h\\ello.c", ROOT_ID)
        self.assertEqual(file.name, "h\\ello.c")

    def test_file_kind_character(self):
        file = inventory.InventoryFile(b"123", "hello.c", ROOT_ID)
        self.assertEqual(file.kind_character(), "")

    def test_dir_kind_character(self):
        dir = inventory.InventoryDirectory(b"123", "hello.c", ROOT_ID)
        self.assertEqual(dir.kind_character(), "/")

    def test_link_kind_character(self):
        dir = inventory.InventoryLink(b"123", "hello.c", ROOT_ID)
        self.assertEqual(dir.kind_character(), "@")

    def test_tree_ref_kind_character(self):
        dir = TreeReference(b"123", "hello.c", ROOT_ID)
        self.assertEqual(dir.kind_character(), "+")

    def test_dir_detect_changes(self):
        left = inventory.InventoryDirectory(b"123", "hello.c", ROOT_ID)
        right = inventory.InventoryDirectory(b"123", "hello.c", ROOT_ID)
        self.assertEqual((False, False), left.detect_changes(right))
        self.assertEqual((False, False), right.detect_changes(left))

    def test_file_detect_changes(self):
        left = inventory.InventoryFile(b"123", "hello.c", ROOT_ID, text_sha1=b"123")
        right = inventory.InventoryFile(b"123", "hello.c", ROOT_ID, text_sha1=b"123")
        self.assertEqual((False, False), left.detect_changes(right))
        self.assertEqual((False, False), right.detect_changes(left))
        left = inventory.InventoryFile(
            b"123", "hello.c", ROOT_ID, text_sha1=b"123", executable=True
        )
        self.assertEqual((False, True), left.detect_changes(right))
        self.assertEqual((False, True), right.detect_changes(left))
        right = inventory.InventoryFile(b"123", "hello.c", ROOT_ID, text_sha1=b"321")
        self.assertEqual((True, True), left.detect_changes(right))
        self.assertEqual((True, True), right.detect_changes(left))

    def test_symlink_detect_changes(self):
        left = inventory.InventoryLink(b"123", "hello.c", ROOT_ID, symlink_target="foo")
        right = inventory.InventoryLink(
            b"123", "hello.c", ROOT_ID, symlink_target="foo"
        )
        self.assertEqual((False, False), left.detect_changes(right))
        self.assertEqual((False, False), right.detect_changes(left))
        left = inventory.InventoryLink(
            b"123", "hello.c", ROOT_ID, symlink_target="different"
        )
        self.assertEqual((True, False), left.detect_changes(right))
        self.assertEqual((True, False), right.detect_changes(left))

    def test_file_has_text(self):
        file = inventory.InventoryFile(b"123", "hello.c", ROOT_ID)
        self.assertTrue(file.has_text())

    def test_directory_has_text(self):
        dir = inventory.InventoryDirectory(b"123", "hello.c", ROOT_ID)
        self.assertFalse(dir.has_text())

    def test_link_has_text(self):
        link = inventory.InventoryLink(b"123", "hello.c", ROOT_ID)
        self.assertFalse(link.has_text())

    def test_make_entry(self):
        self.assertIsInstance(
            inventory.make_entry("file", "name", ROOT_ID), inventory.InventoryFile
        )
        self.assertIsInstance(
            inventory.make_entry("symlink", "name", ROOT_ID), inventory.InventoryLink
        )
        self.assertIsInstance(
            inventory.make_entry("directory", "name", ROOT_ID),
            inventory.InventoryDirectory,
        )

    def test_make_entry_non_normalized(self):
        if osutils.normalizes_filenames():
            entry = inventory.make_entry("file", "a\u030a", ROOT_ID)
            self.assertEqual("\xe5", entry.name)
            self.assertIsInstance(entry, inventory.InventoryFile)
        else:
            self.assertRaises(
                bzrformats_errors.InvalidNormalization,
                inventory.make_entry,
                "file",
                "a\u030a",
                ROOT_ID,
            )


class TestDescribeChanges(TestCase):
    def test_describe_change(self):
        # we need to test the following change combinations:
        # rename
        # reparent
        # modify
        # gone
        # added
        # renamed/reparented and modified
        # change kind (perhaps can't be done yet?)
        # also, merged in combination with all of these?
        old_a = InventoryFile(
            b"a-id", "a_file", ROOT_ID, text_sha1=b"123132", text_size=0
        )
        new_a = InventoryFile(
            b"a-id", "a_file", ROOT_ID, text_sha1=b"123132", text_size=0
        )

        self.assertChangeDescription("unchanged", old_a, new_a)

        new_a = InventoryFile(
            b"a-id", "a_file", ROOT_ID, text_sha1=b"abcabc", text_size=10
        )
        self.assertChangeDescription("modified", old_a, new_a)

        self.assertChangeDescription("added", None, new_a)
        self.assertChangeDescription("removed", old_a, None)
        # perhaps a bit questionable but seems like the most reasonable thing...
        self.assertChangeDescription("unchanged", None, None)

        # in this case it's both renamed and modified; show a rename and
        # modification:
        new_a = InventoryFile(
            b"a-id", "newfilename", ROOT_ID, text_sha1=b"abcabc", text_size=10
        )
        self.assertChangeDescription("modified and renamed", old_a, new_a)

        # reparenting is 'renaming'
        new_a = InventoryFile(
            b"a-id", old_a.name, b"somedir-id", text_sha1=b"abcabc", text_size=10
        )
        self.assertChangeDescription("modified and renamed", old_a, new_a)

        # reset the content values so its not modified
        new_a = InventoryFile(
            b"a-id",
            "newfilename",
            b"somedir-id",
            text_size=old_a.text_size,
            text_sha1=old_a.text_sha1,
        )
        self.assertChangeDescription("renamed", old_a, new_a)

        # reparenting is 'renaming'
        new_a = InventoryFile(
            b"a-id",
            old_a.name,
            b"somedir-id",
            text_size=old_a.text_size,
            text_sha1=old_a.text_sha1,
        )
        self.assertChangeDescription("renamed", old_a, new_a)

    def assertChangeDescription(self, expected_change, old_ie, new_ie):
        change = InventoryEntry.describe_change(old_ie, new_ie)
        self.assertEqual(expected_change, change)


class TestCHKInventory(TestCaseWithMemoryTransport):
    def get_chk_bytes(self):
        factory = groupcompress.make_pack_factory(True, True, 1)
        trans = self.get_transport("")
        return factory(trans)

    def read_bytes(self, chk_bytes, key):
        stream = chk_bytes.get_record_stream([key], "unordered", True)
        return next(stream).get_bytes_as("fulltext")

    def test_deserialise_gives_CHKInventory(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual(b"revid", new_inv.revision_id)
        self.assertEqual("directory", new_inv.root.kind)
        self.assertEqual(inv.root.file_id, new_inv.root.file_id)
        self.assertEqual(inv.root.parent_id, new_inv.root.parent_id)
        self.assertEqual(inv.root.name, new_inv.root.name)
        self.assertEqual(b"rootrev", new_inv.root.revision)
        self.assertEqual(b"plain", new_inv._search_key_name)

    def test_deserialise_wrong_revid(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        self.assertRaises(
            ValueError, CHKInventory.deserialise, chk_bytes, lines, (b"revid2",)
        )

    def test_captures_rev_root_byid(self):
        inv = Inventory(revision_id=b"foo", root_revision=b"bar")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        self.assertEqual(
            [
                b"chkinventory:\n",
                b"revision_id: foo\n",
                b"root_id: TREE_ROOT\n",
                b"parent_id_basename_to_file_id: sha1:eb23f0ad4b07f48e88c76d4c94292be57fb2785f\n",
                b"id_to_entry: sha1:debfe920f1f10e7929260f0534ac9a24d7aabbb4\n",
            ],
            lines,
        )
        chk_inv = CHKInventory.deserialise(chk_bytes, lines, (b"foo",))
        self.assertEqual(b"plain", chk_inv._search_key_name)

    def test_captures_parent_id_basename_index(self):
        inv = Inventory(revision_id=b"foo", root_revision=b"bar")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        self.assertEqual(
            [
                b"chkinventory:\n",
                b"revision_id: foo\n",
                b"root_id: TREE_ROOT\n",
                b"parent_id_basename_to_file_id: sha1:eb23f0ad4b07f48e88c76d4c94292be57fb2785f\n",
                b"id_to_entry: sha1:debfe920f1f10e7929260f0534ac9a24d7aabbb4\n",
            ],
            lines,
        )
        chk_inv = CHKInventory.deserialise(chk_bytes, lines, (b"foo",))
        self.assertEqual(b"plain", chk_inv._search_key_name)

    def test_captures_search_key_name(self):
        inv = Inventory(revision_id=b"foo", root_revision=b"bar")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(
            chk_bytes, inv, search_key_name=b"hash-16-way"
        )
        lines = chk_inv.to_lines()
        self.assertEqual(
            [
                b"chkinventory:\n",
                b"search_key_name: hash-16-way\n",
                b"root_id: TREE_ROOT\n",
                b"parent_id_basename_to_file_id: sha1:eb23f0ad4b07f48e88c76d4c94292be57fb2785f\n",
                b"revision_id: foo\n",
                b"id_to_entry: sha1:debfe920f1f10e7929260f0534ac9a24d7aabbb4\n",
            ],
            lines,
        )
        chk_inv = CHKInventory.deserialise(chk_bytes, lines, (b"foo",))
        self.assertEqual(b"hash-16-way", chk_inv._search_key_name)

    def test_directory_children_on_demand(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        root_entry = new_inv.get_entry(inv.root.file_id)
        self.assertEqual({"file"}, set(inv.get_children(root_entry.file_id)))
        file_direct = new_inv.get_entry(b"fileid")
        file_found = inv.get_children(root_entry.file_id)["file"]
        self.assertEqual(file_direct.kind, file_found.kind)
        self.assertEqual(file_direct.file_id, file_found.file_id)
        self.assertEqual(file_direct.parent_id, file_found.parent_id)
        self.assertEqual(file_direct.name, file_found.name)
        self.assertEqual(file_direct.revision, file_found.revision)
        self.assertEqual(file_direct.text_sha1, file_found.text_sha1)
        self.assertEqual(file_direct.text_size, file_found.text_size)
        self.assertEqual(file_direct.executable, file_found.executable)

    def test_from_inventory_maximum_size(self):
        # from_inventory supports the maximum_size parameter.
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv, 120)
        chk_inv.id_to_entry._ensure_root()
        self.assertEqual(120, chk_inv.id_to_entry._root_node.maximum_size)
        self.assertEqual(1, chk_inv.id_to_entry._root_node._key_width)
        p_id_basename = chk_inv.parent_id_basename_to_file_id
        p_id_basename._ensure_root()
        self.assertEqual(120, p_id_basename._root_node.maximum_size)
        self.assertEqual(2, p_id_basename._root_node._key_width)

    def test_iter_all_ids(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        fileids = sorted(new_inv.iter_all_ids())
        self.assertEqual([inv.root.file_id, b"fileid"], fileids)

    def test__len__(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        self.assertEqual(2, len(chk_inv))

    def test_get_entry(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        root_entry = new_inv.get_entry(inv.root.file_id)
        file_entry = new_inv.get_entry(b"fileid")
        self.assertEqual("directory", root_entry.kind)
        self.assertEqual(inv.root.file_id, root_entry.file_id)
        self.assertEqual(inv.root.parent_id, root_entry.parent_id)
        self.assertEqual(inv.root.name, root_entry.name)
        self.assertEqual(b"rootrev", root_entry.revision)
        self.assertEqual("file", file_entry.kind)
        self.assertEqual(b"fileid", file_entry.file_id)
        self.assertEqual(inv.root.file_id, file_entry.parent_id)
        self.assertEqual("file", file_entry.name)
        self.assertEqual(b"filerev", file_entry.revision)
        self.assertEqual(b"ffff", file_entry.text_sha1)
        self.assertEqual(1, file_entry.text_size)
        self.assertEqual(True, file_entry.executable)
        self.assertRaises(NoSuchId, new_inv.get_entry, "missing")

    def test_has_id_true(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        self.assertTrue(chk_inv.has_id(b"fileid"))
        self.assertTrue(chk_inv.has_id(inv.root.file_id))

    def test_has_id_not(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        self.assertFalse(chk_inv.has_id(b"fileid"))

    def test_id2path(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        direntry = InventoryDirectory(
            b"dirid", "dir", inv.root.file_id, revision=b"filerev"
        )
        fileentry = InventoryFile(
            b"fileid",
            "file",
            b"dirid",
            revision=b"filerev",
            executable=True,
            text_sha1=b"ffff",
            text_size=1,
        )

        inv.add(direntry)
        inv.add(fileentry)
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual("", new_inv.id2path(inv.root.file_id))
        self.assertEqual("dir", new_inv.id2path(b"dirid"))
        self.assertEqual("dir/file", new_inv.id2path(b"fileid"))

    def test_path2id(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        direntry = InventoryDirectory(
            b"dirid", "dir", inv.root.file_id, revision=b"filerev"
        )

        fileentry = InventoryFile(
            b"fileid",
            "file",
            b"dirid",
            revision=b"filerev",
            executable=True,
            text_sha1=b"ffff",
            text_size=1,
        )
        inv.add(direntry)
        inv.add(fileentry)
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual(inv.root.file_id, new_inv.path2id(""))
        self.assertEqual(b"dirid", new_inv.path2id("dir"))
        self.assertEqual(b"fileid", new_inv.path2id("dir/file"))

    def test_create_by_apply_delta_sets_root(self):
        inv = Inventory(root_revision=b"myrootrev", revision_id=b"revid")
        chk_bytes = self.get_chk_bytes()
        base_inv = CHKInventory.from_inventory(chk_bytes, inv)
        inv.revision_id = b"expectedid"
        inv.add_path("", "directory", b"myrootid", revision=b"myrootrev")
        reference_inv = CHKInventory.from_inventory(chk_bytes, inv)
        delta = InventoryDelta(
            [("", None, base_inv.root.file_id, None), (None, "", b"myrootid", inv.root)]
        )
        new_inv = base_inv.create_by_apply_delta(delta, b"expectedid")
        self.assertEqual(reference_inv.root, new_inv.root)

    def test_create_by_apply_delta_empty_add_child(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        chk_bytes = self.get_chk_bytes()
        base_inv = CHKInventory.from_inventory(chk_bytes, inv)
        a_entry = InventoryFile(
            b"A-id",
            "A",
            inv.root.file_id,
            revision=b"filerev",
            executable=True,
            text_sha1=b"ffff",
            text_size=1,
        )
        inv.add(a_entry)
        inv.revision_id = b"expectedid"
        reference_inv = CHKInventory.from_inventory(chk_bytes, inv)
        delta = InventoryDelta([(None, "A", b"A-id", a_entry)])
        new_inv = base_inv.create_by_apply_delta(delta, b"expectedid")
        # new_inv should be the same as reference_inv.
        self.assertEqual(reference_inv.revision_id, new_inv.revision_id)
        self.assertEqual(reference_inv.root_id, new_inv.root_id)
        reference_inv.id_to_entry._ensure_root()
        new_inv.id_to_entry._ensure_root()
        self.assertEqual(
            reference_inv.id_to_entry._root_node._key,
            new_inv.id_to_entry._root_node._key,
        )

    def test_create_by_apply_delta_empty_add_child_updates_parent_id(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        chk_bytes = self.get_chk_bytes()
        base_inv = CHKInventory.from_inventory(chk_bytes, inv)
        a_entry = InventoryFile(
            b"A-id",
            "A",
            inv.root.file_id,
            revision=b"filerev",
            executable=True,
            text_sha1=b"ffff",
            text_size=1,
        )
        inv.add(a_entry)
        inv.revision_id = b"expectedid"
        reference_inv = CHKInventory.from_inventory(chk_bytes, inv)
        delta = InventoryDelta([(None, "A", b"A-id", a_entry)])
        new_inv = base_inv.create_by_apply_delta(delta, b"expectedid")
        reference_inv.id_to_entry._ensure_root()
        reference_inv.parent_id_basename_to_file_id._ensure_root()
        new_inv.id_to_entry._ensure_root()
        new_inv.parent_id_basename_to_file_id._ensure_root()
        # new_inv should be the same as reference_inv.
        self.assertEqual(reference_inv.revision_id, new_inv.revision_id)
        self.assertEqual(reference_inv.root_id, new_inv.root_id)
        self.assertEqual(
            reference_inv.id_to_entry._root_node._key,
            new_inv.id_to_entry._root_node._key,
        )
        self.assertEqual(
            reference_inv.parent_id_basename_to_file_id._root_node._key,
            new_inv.parent_id_basename_to_file_id._root_node._key,
        )

    def test_iter_changes(self):
        # Low level bootstrapping smoke test; comprehensive generic tests via
        # InterTree are coming.
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        inv2 = Inventory(revision_id=b"revid2", root_revision=b"rootrev")
        inv2.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev2",
                executable=False,
                text_sha1=b"bbbb",
                text_size=2,
            )
        )
        # get fresh objects.
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        inv_1 = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        chk_inv2 = CHKInventory.from_inventory(chk_bytes, inv2)
        lines = chk_inv2.to_lines()
        inv_2 = CHKInventory.deserialise(chk_bytes, lines, (b"revid2",))
        self.assertEqual(
            [
                (
                    b"fileid",
                    ("file", "file"),
                    True,
                    (True, True),
                    (b"TREE_ROOT", b"TREE_ROOT"),
                    ("file", "file"),
                    ("file", "file"),
                    (False, True),
                )
            ],
            list(inv_1.iter_changes(inv_2)),
        )

    def test_parent_id_basename_to_file_id_index_enabled(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                inv.root.file_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        # get fresh objects.
        chk_bytes = self.get_chk_bytes()
        tmp_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = tmp_inv.to_lines()
        chk_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertIsInstance(chk_inv.parent_id_basename_to_file_id, chk_map.CHKMap)
        self.assertEqual(
            {(b"", b""): b"TREE_ROOT", (b"TREE_ROOT", b"file"): b"fileid"},
            dict(chk_inv.parent_id_basename_to_file_id.iteritems()),
        )

    def test_file_entry_to_bytes(self):
        CHKInventory(None)
        ie = inventory.InventoryFile(
            b"file-id",
            "filename",
            b"parent-id",
            executable=True,
            revision=b"file-rev-id",
            text_sha1=b"abcdefgh",
            text_size=100,
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(
            b"file: file-id\nparent-id\nfilename\nfile-rev-id\nabcdefgh\n100\nY",
            bytes,
        )
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual(
            (b"filename", b"file-id", b"file-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def test_file2_entry_to_bytes(self):
        CHKInventory(None)
        # \u30a9 == 'omega'
        ie = inventory.InventoryFile(
            b"file-id",
            "\u03a9name",
            b"parent-id",
            executable=False,
            revision=b"file-rev-id",
            text_sha1=b"123456",
            text_size=25,
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(
            b"file: file-id\nparent-id\n\xce\xa9name\nfile-rev-id\n123456\n25\nN",
            bytes,
        )
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual(
            (b"\xce\xa9name", b"file-id", b"file-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def test_dir_entry_to_bytes(self):
        CHKInventory(None)
        ie = inventory.InventoryDirectory(
            b"dir-id", "dirname", b"parent-id", revision=b"dir-rev-id"
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(b"dir: dir-id\nparent-id\ndirname\ndir-rev-id", bytes)
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual(
            (b"dirname", b"dir-id", b"dir-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def test_dir2_entry_to_bytes(self):
        CHKInventory(None)
        ie = inventory.InventoryDirectory(
            b"dir-id", "dir\u03a9name", b"pid", revision=b"dir-rev-id"
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(b"dir: dir-id\npid\ndir\xce\xa9name\ndir-rev-id", bytes)
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual(b"pid", ie2.parent_id)
        self.assertEqual(
            (b"dir\xce\xa9name", b"dir-id", b"dir-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def test_symlink_entry_to_bytes(self):
        CHKInventory(None)
        ie = inventory.InventoryLink(
            b"link-id",
            "linkname",
            b"parent-id",
            revision=b"link-rev-id",
            symlink_target="target/path",
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(
            b"symlink: link-id\nparent-id\nlinkname\nlink-rev-id\ntarget/path",
            bytes,
        )
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertIsInstance(ie2.symlink_target, str)
        self.assertEqual(
            (b"linkname", b"link-id", b"link-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def test_symlink2_entry_to_bytes(self):
        CHKInventory(None)
        ie = inventory.InventoryLink(
            b"link-id",
            "link\u03a9name",
            b"parent-id",
            revision=b"link-rev-id",
            symlink_target="target/\u03a9path",
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(
            b"symlink: link-id\nparent-id\nlink\xce\xa9name\n"
            b"link-rev-id\ntarget/\xce\xa9path",
            bytes,
        )
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertIsInstance(ie2.symlink_target, str)
        self.assertEqual(
            (b"link\xce\xa9name", b"link-id", b"link-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def test_tree_reference_entry_to_bytes(self):
        CHKInventory(None)
        ie = inventory.TreeReference(
            b"tree-root-id",
            "tree\u03a9name",
            b"parent-id",
            revision=b"tree-rev-id",
            reference_revision=b"ref-rev-id",
        )
        bytes = _chk_inventory_entry_to_bytes(ie)
        self.assertEqual(
            b"tree: tree-root-id\nparent-id\ntree\xce\xa9name\ntree-rev-id\nref-rev-id",
            bytes,
        )
        ie2 = _chk_inventory_bytes_to_entry(bytes)
        self.assertEqual(ie, ie2)
        self.assertIsInstance(ie2.name, str)
        self.assertEqual(
            (b"tree\xce\xa9name", b"tree-root-id", b"tree-rev-id"),
            chk_inventory_bytes_to_utf8name_key(bytes),
        )

    def make_basic_utf8_inventory(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        root_id = inv.root.file_id
        inv.add(
            InventoryFile(
                b"fileid",
                "f\xefle",
                root_id,
                revision=b"filerev",
                text_sha1=b"ffff",
                text_size=0,
            )
        )
        inv.add(
            InventoryDirectory(
                b"dirid", "dir-\N{EURO SIGN}", root_id, revision=b"dirrev"
            )
        )
        inv.add(
            InventoryFile(
                b"childid",
                "ch\xefld",
                b"dirid",
                revision=b"filerev",
                text_sha1=b"ffff",
                text_size=0,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        return CHKInventory.deserialise(chk_bytes, lines, (b"revid",))

    def test__preload_handles_utf8(self):
        new_inv = self.make_basic_utf8_inventory()
        self.assertEqual({}, new_inv._fileid_to_entry_cache)
        self.assertFalse(new_inv._fully_cached)
        new_inv._preload_cache()
        self.assertEqual(
            sorted([new_inv.root_id, b"fileid", b"dirid", b"childid"]),
            sorted(new_inv._fileid_to_entry_cache.keys()),
        )
        ie_root = new_inv._fileid_to_entry_cache[new_inv.root_id]
        self.assertEqual(
            ["dir-\N{EURO SIGN}", "f\xefle"],
            [ie.name for ie in new_inv.iter_sorted_children(ie_root.file_id)],
        )
        ie_dir = new_inv._fileid_to_entry_cache[b"dirid"]
        self.assertEqual(
            ["ch\xefld"],
            [ie.name for ie in new_inv.iter_sorted_children(ie_dir.file_id)],
        )

    def test__preload_populates_cache(self):
        inv = Inventory(revision_id=b"revid", root_revision=b"rootrev")
        root_id = inv.root.file_id
        inv.add(
            InventoryFile(
                b"fileid",
                "file",
                root_id,
                revision=b"filerev",
                executable=True,
                text_sha1=b"ffff",
                text_size=1,
            )
        )
        inv.add(InventoryDirectory(b"dirid", "dir", root_id, revision=b"dirrev"))
        inv.add(
            InventoryFile(
                b"childid",
                "child",
                b"dirid",
                revision=b"filerev",
                executable=False,
                text_sha1=b"dddd",
                text_size=1,
            )
        )
        chk_bytes = self.get_chk_bytes()
        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
        lines = chk_inv.to_lines()
        new_inv = CHKInventory.deserialise(chk_bytes, lines, (b"revid",))
        self.assertEqual({}, new_inv._fileid_to_entry_cache)
        self.assertFalse(new_inv._fully_cached)
        new_inv._preload_cache()
        self.assertEqual(
            sorted([root_id, b"fileid", b"dirid", b"childid"]),
            sorted(new_inv._fileid_to_entry_cache.keys()),
        )
        self.assertTrue(new_inv._fully_cached)
        ie_root = new_inv._fileid_to_entry_cache[root_id]
        self.assertEqual(
            ["dir", "file"],
            [ie.name for ie in new_inv.iter_sorted_children(ie_root.file_id)],
        )
        ie_dir = new_inv._fileid_to_entry_cache[b"dirid"]
        self.assertEqual(
            ["child"], [ie.name for ie in new_inv.iter_sorted_children(ie_dir.file_id)]
        )

    def test__preload_handles_partially_evaluated_inventory(self):
        new_inv = self.make_basic_utf8_inventory()
        ie = new_inv.get_entry(new_inv.root_id)
        self.assertEqual(
            ["dir-\N{EURO SIGN}", "f\xefle"],
            [c.name for c in new_inv.iter_sorted_children(ie.file_id)],
        )
        new_inv._preload_cache()
        # No change
        self.assertEqual(
            ["dir-\N{EURO SIGN}", "f\xefle"],
            [c.name for c in new_inv.iter_sorted_children(ie.file_id)],
        )
        self.assertEqual(
            ["ch\xefld"], [c.name for c in new_inv.iter_sorted_children(b"dirid")]
        )

    def test_filter_change_in_renamed_subfolder(self):
        inv = Inventory(b"tree-root", root_revision=b"rootrev")
        src_ie = inv.add_path("src", "directory", b"src-id", revision=b"srcrev")
        inv.add_path("src/sub/", "directory", b"sub-id", revision=b"subrev")
        a_ie = inv.add_path(
            "src/sub/a",
            "file",
            b"a-id",
            revision=b"filerev",
            text_sha1=osutils.sha_string(b"content\n"),
            text_size=len(b"content\n"),
        )
        chk_bytes = self.get_chk_bytes()
        inv = CHKInventory.from_inventory(chk_bytes, inv)
        inv = inv.create_by_apply_delta(
            InventoryDelta(
                [
                    ("src/sub/a", "src/sub/a", b"a-id", a_ie),
                    ("src", "src2", b"src-id", src_ie),
                ]
            ),
            b"new-rev-2",
        )
        new_inv = inv.filter([b"a-id", b"src-id"])
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("src", b"src-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )


class TestCHKInventoryExpand(TestCaseWithMemoryTransport):
    def get_chk_bytes(self):
        factory = groupcompress.make_pack_factory(True, True, 1)
        trans = self.get_transport("")
        return factory(trans)

    def make_dir(self, inv, name, parent_id, revision):
        ie = inv.make_entry(
            "directory",
            name,
            parent_id,
            name.encode("utf-8") + b"-id",
            revision=revision,
        )
        inv.add(ie)

    def make_file(self, inv, name, parent_id, revision, content=b"content\n"):
        ie = inv.make_entry(
            "file",
            name,
            parent_id,
            name.encode("utf-8") + b"-id",
            text_sha1=osutils.sha_string(content),
            text_size=len(content),
            revision=revision,
        )
        inv.add(ie)

    def make_simple_inventory(self):
        inv = Inventory(b"TREE_ROOT", revision_id=b"revid", root_revision=b"rootrev")
        # /                 TREE_ROOT
        # dir1/             dir1-id
        #   sub-file1       sub-file1-id
        #   sub-file2       sub-file2-id
        #   sub-dir1/       sub-dir1-id
        #     subsub-file1  subsub-file1-id
        # dir2/             dir2-id
        #   sub2-file1      sub2-file1-id
        # top               top-id
        self.make_dir(inv, "dir1", b"TREE_ROOT", b"dirrev")
        self.make_dir(inv, "dir2", b"TREE_ROOT", b"dirrev")
        self.make_dir(inv, "sub-dir1", b"dir1-id", b"dirrev")
        self.make_file(inv, "top", b"TREE_ROOT", b"filerev")
        self.make_file(inv, "sub-file1", b"dir1-id", b"filerev")
        self.make_file(inv, "sub-file2", b"dir1-id", b"filerev")
        self.make_file(inv, "subsub-file1", b"sub-dir1-id", b"filerev")
        self.make_file(inv, "sub2-file1", b"dir2-id", b"filerev")
        chk_bytes = self.get_chk_bytes()
        #  use a small maximum_size to force internal paging structures
        chk_inv = CHKInventory.from_inventory(
            chk_bytes, inv, maximum_size=100, search_key_name=b"hash-255-way"
        )
        lines = chk_inv.to_lines()
        return CHKInventory.deserialise(chk_bytes, lines, (b"revid",))

    def assert_Getitems(self, expected_fileids, inv, file_ids):
        self.assertEqual(
            sorted(expected_fileids),
            sorted([ie.file_id for ie in inv._getitems(file_ids)]),
        )

    def assertExpand(self, all_ids, inv, file_ids):
        (val_all_ids, val_children) = inv._expand_fileids_to_parents_and_children(
            file_ids
        )
        self.assertEqual(set(all_ids), val_all_ids)
        entries = inv._getitems(val_all_ids)
        expected_children = {}
        for entry in entries:
            s = expected_children.setdefault(entry.parent_id, [])
            s.append(entry.file_id)
        val_children = {k: sorted(v) for k, v in val_children.items()}
        expected_children = {k: sorted(v) for k, v in expected_children.items()}
        self.assertEqual(expected_children, val_children)

    def test_make_simple_inventory(self):
        inv = self.make_simple_inventory()
        layout = []
        for path, entry in inv.iter_entries_by_dir():
            layout.append((path, entry.file_id))
        self.assertEqual(
            [
                ("", b"TREE_ROOT"),
                ("dir1", b"dir1-id"),
                ("dir2", b"dir2-id"),
                ("top", b"top-id"),
                ("dir1/sub-dir1", b"sub-dir1-id"),
                ("dir1/sub-file1", b"sub-file1-id"),
                ("dir1/sub-file2", b"sub-file2-id"),
                ("dir1/sub-dir1/subsub-file1", b"subsub-file1-id"),
                ("dir2/sub2-file1", b"sub2-file1-id"),
            ],
            layout,
        )

    def test__getitems(self):
        inv = self.make_simple_inventory()
        # Reading from disk
        self.assert_Getitems([b"dir1-id"], inv, [b"dir1-id"])
        self.assertIn(b"dir1-id", inv._fileid_to_entry_cache)
        self.assertNotIn(b"sub-file2-id", inv._fileid_to_entry_cache)
        # From cache
        self.assert_Getitems([b"dir1-id"], inv, [b"dir1-id"])
        # Mixed
        self.assert_Getitems(
            [b"dir1-id", b"sub-file2-id"], inv, [b"dir1-id", b"sub-file2-id"]
        )
        self.assertIn(b"dir1-id", inv._fileid_to_entry_cache)
        self.assertIn(b"sub-file2-id", inv._fileid_to_entry_cache)

    def test_single_file(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b"TREE_ROOT", b"top-id"], inv, [b"top-id"])

    def test_get_all_parents(self):
        inv = self.make_simple_inventory()
        self.assertExpand(
            [
                b"TREE_ROOT",
                b"dir1-id",
                b"sub-dir1-id",
                b"subsub-file1-id",
            ],
            inv,
            [b"subsub-file1-id"],
        )

    def test_get_children(self):
        inv = self.make_simple_inventory()
        self.assertExpand(
            [
                b"TREE_ROOT",
                b"dir1-id",
                b"sub-dir1-id",
                b"sub-file1-id",
                b"sub-file2-id",
                b"subsub-file1-id",
            ],
            inv,
            [b"dir1-id"],
        )

    def test_from_root(self):
        inv = self.make_simple_inventory()
        self.assertExpand(
            [
                b"TREE_ROOT",
                b"dir1-id",
                b"dir2-id",
                b"sub-dir1-id",
                b"sub-file1-id",
                b"sub-file2-id",
                b"sub2-file1-id",
                b"subsub-file1-id",
                b"top-id",
            ],
            inv,
            [b"TREE_ROOT"],
        )

    def test_top_level_file(self):
        inv = self.make_simple_inventory()
        self.assertExpand([b"TREE_ROOT", b"top-id"], inv, [b"top-id"])

    def test_subsub_file(self):
        inv = self.make_simple_inventory()
        self.assertExpand(
            [b"TREE_ROOT", b"dir1-id", b"sub-dir1-id", b"subsub-file1-id"],
            inv,
            [b"subsub-file1-id"],
        )

    def test_sub_and_root(self):
        inv = self.make_simple_inventory()
        self.assertExpand(
            [b"TREE_ROOT", b"dir1-id", b"sub-dir1-id", b"top-id", b"subsub-file1-id"],
            inv,
            [b"top-id", b"subsub-file1-id"],
        )


class ErrorTests(TestCase):
    def test_duplicate_file_id(self):
        error = DuplicateFileId("a_file_id", "foo")
        self.assertEqualDiff(
            "File id {a_file_id} already exists in inventory as foo", str(error)
        )



# Copyright (C) 2005 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""RIO-Patch format handling for email-safe stanza representation.

This module provides functions to convert between RIO stanzas and RIO-Patch format,
which is designed to be emailed as part of a patch. The format resists common forms
of damage such as newline conversion or removal of trailing whitespace.

The RIO (restricted/reproducible/rfc822-like) format stores data as a series of
stanzas containing fields identified by ASCII names with Unicode or string contents.
"""

# \subsection{\emph{rio} - simple text metaformat}
#
# \emph{r} stands for `restricted', `reproducible', or `rfc822-like'.
#
# The stored data consists of a series of \emph{stanzas}, each of which contains
# \emph{fields} identified by an ascii name, with Unicode or string contents.
# The field tag is constrained to alphanumeric characters.
# There may be more than one field in a stanza with the same name.
#
# The format itself does not deal with character encoding issues, though
# the result will normally be written in Unicode.
#
# The format is intended to be simple enough that there is exactly one character
# stream representation of an object and vice versa, and that this relation
# will continue to hold for future versions of bzr.

import re

from . import rio


def to_patch_lines(stanza, max_width=72):
    """Convert a stanza into RIO-Patch format lines.

    RIO-Patch is a RIO variant designed to be e-mailed as part of a patch.
    It resists common forms of damage such as newline conversion or the removal
    of trailing whitespace, yet is also reasonably easy to read.

    :param max_width: The maximum number of characters per physical line.
    :return: a list of lines
    """
    if max_width <= 6:
        raise ValueError(max_width)
    max_rio_width = max_width - 4
    lines = []
    for pline in stanza.to_lines():
        for line in pline.split(b"\n")[:-1]:
            line = re.sub(b"\\\\", b"\\\\\\\\", line)
            while len(line) > 0:
                partline = line[:max_rio_width]
                line = line[max_rio_width:]
                if len(line) > 0 and line[:1] != [b" "]:
                    break_index = -1
                    break_index = partline.rfind(b" ", -20)
                    if break_index < 3:
                        break_index = partline.rfind(b"-", -20)
                        break_index += 1
                    if break_index < 3:
                        break_index = partline.rfind(b"/", -20)
                    if break_index >= 3:
                        line = partline[break_index:] + line
                        partline = partline[:break_index]
                if len(line) > 0:
                    line = b"  " + line
                partline = re.sub(b"\r", b"\\\\r", partline)
                blank_line = False
                if len(line) > 0:
                    partline += b"\\"
                elif re.search(b" $", partline):
                    partline += b"\\"
                    blank_line = True
                lines.append(b"# " + partline + b"\n")
                if blank_line:
                    lines.append(b"#   \n")
    return lines


def _patch_stanza_iter(line_iter):
    map = {b"\\\\": b"\\", b"\\r": b"\r", b"\\\n": b""}

    def mapget(match):
        return map[match.group(0)]

    last_line = None
    for line in line_iter:
        if line.startswith(b"# "):
            line = line[2:]
        elif line.startswith(b"#"):
            line = line[1:]
        else:
            raise ValueError(f"bad line {line!r}")
        if last_line is not None and len(line) > 2:
            line = line[2:]
        line = re.sub(b"\r", b"", line)
        line = re.sub(b"\\\\(.|\n)", mapget, line)
        if last_line is None:
            last_line = line
        else:
            last_line += line
        if last_line[-1:] == b"\n":
            yield last_line
            last_line = None
    if last_line is not None:
        yield last_line


def read_patch_stanza(line_iter):
    """Convert an iterable of RIO-Patch format lines into a Stanza.

    RIO-Patch is a RIO variant designed to be e-mailed as part of a patch.
    It resists common forms of damage such as newline conversion or the removal
    of trailing whitespace, yet is also reasonably easy to read.

    :return: a Stanza
    """
    return rio.read_stanza(_patch_stanza_iter(line_iter))



# Copyright (C) 2025 Breezy Contributors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Minimal progress bar protocol for bzrformats."""

from typing import Protocol, runtime_checkable


@runtime_checkable
class ProgressBar(Protocol):
    """Protocol for progress reporting."""

    def update(
        self,
        msg: str | None = None,
        current: int | None = None,
        total: int | None = None,
    ) -> None:
        """Report progress.

        :param msg: Description of the current step.
        :param current: Current step number.
        :param total: Total number of steps.
        """
        ...

    def tick(self) -> None:
        """Indicate that some work was done without specific progress info."""
        ...

    def finished(self) -> None:
        """Mark the progress bar as complete."""
        ...



#! /usr/bin/env python3

"""Installation script for bzrformats.

Run it with
 './setup.py install', or
 './setup.py --help' for more options.
"""

import os
import os.path
import sys

try:
    import setuptools  # noqa: F401
except ModuleNotFoundError as e:
    sys.stderr.write(f"[ERROR] Please install setuptools ({e})\n")
    sys.exit(1)

try:
    from setuptools_rust import Binding, RustExtension
except ModuleNotFoundError as e:
    sys.stderr.write(f"[ERROR] Please install setuptools_rust ({e})\n")
    sys.exit(1)

from setuptools import setup

try:
    from packaging.version import Version
except ImportError:
    from distutils.version import LooseVersion as Version

from distutils.command.build_scripts import build_scripts

from setuptools import Command

###############################
# Overridden distutils actions
###############################


class brz_build_scripts(build_scripts):
    """Custom build_scripts command that handles Rust extension binaries.

    This class extends the standard build_scripts command to properly handle
    Rust extension binaries by moving executable Rust extensions from the
    build_lib directory to the scripts directory.
    """

    def run(self):
        """Execute the build_scripts command and handle Rust executables.

        First runs the standard build_scripts process, then moves any Rust
        executable extensions from the build_lib directory to the scripts
        build directory.
        """
        build_scripts.run(self)

        self.run_command("build_ext")
        build_ext = self.get_finalized_command("build_ext")

        for ext in self.distribution.rust_extensions:
            if ext.binding == Binding.Exec:
                # GZ 2021-08-19: Not handling multiple binaries yet.
                os.replace(
                    os.path.join(build_ext.build_lib, ext.name),
                    os.path.join(self.build_dir, ext.name),
                )


class build_man(Command):
    """Custom command to generate the brz.1 manual page.

    This command builds the Breezy extension modules and then uses the
    generate_docs tool to create the brz.1 manual page from the built
    modules.
    """

    def initialize_options(self):
        """Initialize command options.

        No options to initialize for this command.
        """
        pass

    def finalize_options(self):
        """Finalize command options.

        No options to finalize for this command.
        """
        pass

    def run(self):
        """Execute the manual page generation.

        Builds the extension modules, adds the build directory to sys.path,
        and then imports and runs the generate_docs tool to create the
        brz.1 manual page.
        """
        build_ext_cmd = self.get_finalized_command("build_ext")
        build_lib_dir = build_ext_cmd.build_lib
        sys.path.insert(0, os.path.abspath(build_lib_dir))
        import importlib

        importlib.invalidate_caches()
        del sys.modules["breezy"]
        from tools import generate_docs

        generate_docs.main(["generate-docs", "man"])


########################
## Setup
########################

command_classes = {
    "build_man": build_man,
}

from distutils.extension import Extension

ext_modules = []
try:
    from Cython.Compiler.Version import version as cython_version
    from Cython.Distutils import build_ext
except ModuleNotFoundError:
    have_cython = False
    # try to build the extension from the prior generated source.
    print("")
    print(
        "The python package 'Cython' is not available. If the .c files are available,"
    )
    print("they will be built, but modifying the .pyx files will not rebuild them.")
    print("")
    from distutils.command.build_ext import build_ext
else:
    minimum_cython_version = "0.29"
    cython_version_info = Version(cython_version)
    if cython_version_info < Version(minimum_cython_version):
        print(
            "Version of Cython is too old. "
            f"Current is {cython_version}, need at least {minimum_cython_version}."
        )
        print(
            "If the .c files are available, they will be built,"
            " but modifying the .pyx files will not rebuild them."
        )
        have_cython = False
    else:
        have_cython = True


# Override the build_ext if we have Cython available
command_classes["build_ext"] = build_ext
unavailable_files = []


def add_cython_extension(module_name, libraries=None, extra_source=None):
    """Add a Cython extension module to the build configuration.

    This function configures a Cython extension for building. If Cython is
    available, it will compile from .pyx files. Otherwise, it falls back to
    pre-generated .c files. If neither is available, the extension is skipped
    with a warning.

    Args:
        module_name (str): The python path to the module (e.g., 'bzrformats.foo').
            This determines the .pyx and .c file paths to use.
        libraries (list, optional): List of libraries to link against.
            Defaults to None.
        extra_source (list, optional): Additional source files to include.
            Defaults to None.

    Note:
        On Windows, the WIN32 macro is automatically defined for Cython
        compatibility. The function adds appropriate include directories
        and handles the optional nature of extensions for CI builds.
    """
    if extra_source is None:
        extra_source = []
    path = module_name.replace(".", "/")
    cython_name = path + ".pyx"
    c_name = path + ".c"
    define_macros = []
    if sys.platform == "win32":
        # cython uses the macro WIN32 to detect the platform, even though it
        # should be using something like _WIN32 or MS_WINDOWS, oh well, we can
        # give it the right value.
        define_macros.append(("WIN32", None))
    if have_cython:
        source = [cython_name]
    else:
        if not os.path.isfile(c_name):
            unavailable_files.append(c_name)
            return
        else:
            source = [c_name]
    source.extend(extra_source)
    include_dirs = ["breezy"]
    ext_modules.append(
        Extension(
            module_name,
            source,
            define_macros=define_macros,
            libraries=libraries,
            include_dirs=include_dirs,
            optional=os.environ.get("CIBUILDWHEEL", "0") != "1",
        )
    )


add_cython_extension(
    "bzrformats._groupcompress_pyx", extra_source=["bzrformats/diff-delta.c"]
)
add_cython_extension("bzrformats._knit_load_data_pyx")
if sys.platform == "win32":
    add_cython_extension("bzrformats._dirstate_helpers_pyx", libraries=["Ws2_32"])
else:
    add_cython_extension("bzrformats._dirstate_helpers_pyx")
add_cython_extension("bzrformats._btree_serializer_pyx")


if unavailable_files:
    print("C extension(s) not found:")
    print("   {}".format("\n  ".join(unavailable_files)))
    print("The python versions will be used instead.")
    print("")


import site

site.ENABLE_USER_SITE = "--user" in sys.argv

rust_extensions = [
    RustExtension(
        "bzrformats._bzr_rs", "crates/bazaar-py/Cargo.toml", binding=Binding.PyO3
    ),
    RustExtension(
        "bzrformats._osutils_rs", "crates/osutils-py/Cargo.toml", binding=Binding.PyO3
    ),
]
entry_points = {}

# std setup
setup(
    cmdclass=command_classes,
    ext_modules=ext_modules,
    entry_points=entry_points,
    rust_extensions=rust_extensions,
)



# Copyright (C) 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for bisect_multi."""

from ..bisect_multi import bisect_multi_bytes
from . import TestCase


class TestBisectMultiBytes(TestCase):
    def test_lookup_no_keys_no_calls(self):
        calls = []

        def missing_content(location_keys):
            calls.append(location_keys)
            return ((location_key, False) for location_key in location_keys)

        self.assertEqual([], list(bisect_multi_bytes(missing_content, 100, [])))
        self.assertEqual([], calls)

    def test_lookup_missing_key_no_content(self):
        """Doing a lookup in a zero-length file still does a single request.

        This makes sense because the bisector cannot tell how long content is
        and its more flexible to only stop when the content object says 'False'
        for a given location, key pair.
        """
        calls = []

        def missing_content(location_keys):
            calls.append(location_keys)
            return ((location_key, False) for location_key in location_keys)

        self.assertEqual(
            [], list(bisect_multi_bytes(missing_content, 0, ["foo", "bar"]))
        )
        self.assertEqual([[(0, "foo"), (0, "bar")]], calls)

    def test_lookup_missing_key_before_all_others(self):
        calls = []

        def missing_first_content(location_keys):
            # returns -1 for all keys unless the byte offset is 0 when it
            # returns False
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[0] == 0:
                    result.append((location_key, False))
                else:
                    result.append((location_key, -1))
            return result

        # given a 0 length file, this should terminate with one call.
        self.assertEqual(
            [], list(bisect_multi_bytes(missing_first_content, 0, ["foo", "bar"]))
        )
        self.assertEqual([[(0, "foo"), (0, "bar")]], calls)
        del calls[:]
        # given a 2 length file, this should make two calls - 1, 0.
        self.assertEqual(
            [], list(bisect_multi_bytes(missing_first_content, 2, ["foo", "bar"]))
        )
        self.assertEqual(
            [
                [(1, "foo"), (1, "bar")],
                [(0, "foo"), (0, "bar")],
            ],
            calls,
        )
        del calls[:]
        # given a really long file - 200MB, this should make a series of calls with the
        # gap between adjactent calls dropping by 50% each time. We choose a
        # length which just under a power of two to generate a corner case in
        # bisection - naively using power of two reduction in size can lead to
        # a very long tail in the bisection process. The current users of
        # the bisect_multi_bytes api are not expected to be concerned by this,
        # as the delta gets down to 4K (the minimum we expect to read and
        # parse) within 16 steps even on a 200MB index (which at 4 keys/K is
        # 800 thousand keys, and log2 of 800000 is 19 - so we're doing log2
        # steps in the worst case there.
        self.assertEqual(
            [],
            list(
                bisect_multi_bytes(missing_first_content, 268435456 - 1, ["foo", "bar"])
            ),
        )
        self.assertEqual(
            [
                [(134217727, "foo"), (134217727, "bar")],
                [(67108864, "foo"), (67108864, "bar")],
                [(33554433, "foo"), (33554433, "bar")],
                [(16777218, "foo"), (16777218, "bar")],
                [(8388611, "foo"), (8388611, "bar")],
                [(4194308, "foo"), (4194308, "bar")],
                [(2097157, "foo"), (2097157, "bar")],
                [(1048582, "foo"), (1048582, "bar")],
                [(524295, "foo"), (524295, "bar")],
                [(262152, "foo"), (262152, "bar")],
                [(131081, "foo"), (131081, "bar")],
                [(65546, "foo"), (65546, "bar")],
                [(32779, "foo"), (32779, "bar")],
                [(16396, "foo"), (16396, "bar")],
                [(8205, "foo"), (8205, "bar")],
                [(4110, "foo"), (4110, "bar")],
                [(2063, "foo"), (2063, "bar")],
                [(1040, "foo"), (1040, "bar")],
                [(529, "foo"), (529, "bar")],
                [(274, "foo"), (274, "bar")],
                [(147, "foo"), (147, "bar")],
                [(84, "foo"), (84, "bar")],
                [(53, "foo"), (53, "bar")],
                [(38, "foo"), (38, "bar")],
                [(31, "foo"), (31, "bar")],
                [(28, "foo"), (28, "bar")],
                [(27, "foo"), (27, "bar")],
                [(26, "foo"), (26, "bar")],
                [(25, "foo"), (25, "bar")],
                [(24, "foo"), (24, "bar")],
                [(23, "foo"), (23, "bar")],
                [(22, "foo"), (22, "bar")],
                [(21, "foo"), (21, "bar")],
                [(20, "foo"), (20, "bar")],
                [(19, "foo"), (19, "bar")],
                [(18, "foo"), (18, "bar")],
                [(17, "foo"), (17, "bar")],
                [(16, "foo"), (16, "bar")],
                [(15, "foo"), (15, "bar")],
                [(14, "foo"), (14, "bar")],
                [(13, "foo"), (13, "bar")],
                [(12, "foo"), (12, "bar")],
                [(11, "foo"), (11, "bar")],
                [(10, "foo"), (10, "bar")],
                [(9, "foo"), (9, "bar")],
                [(8, "foo"), (8, "bar")],
                [(7, "foo"), (7, "bar")],
                [(6, "foo"), (6, "bar")],
                [(5, "foo"), (5, "bar")],
                [(4, "foo"), (4, "bar")],
                [(3, "foo"), (3, "bar")],
                [(2, "foo"), (2, "bar")],
                [(1, "foo"), (1, "bar")],
                [(0, "foo"), (0, "bar")],
            ],
            calls,
        )

    def test_lookup_missing_key_after_all_others(self):
        calls = []
        end = None

        def missing_last_content(location_keys):
            # returns +1 for all keys unless the byte offset is 'end' when it
            # returns False
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[0] == end:
                    result.append((location_key, False))
                else:
                    result.append((location_key, +1))
            return result

        # given a 0 length file, this should terminate with one call.
        end = 0
        self.assertEqual(
            [], list(bisect_multi_bytes(missing_last_content, 0, ["foo", "bar"]))
        )
        self.assertEqual([[(0, "foo"), (0, "bar")]], calls)
        del calls[:]
        end = 2
        # given a 3 length file, this should make two calls - 1, 2.
        self.assertEqual(
            [], list(bisect_multi_bytes(missing_last_content, 3, ["foo", "bar"]))
        )
        self.assertEqual(
            [
                [(1, "foo"), (1, "bar")],
                [(2, "foo"), (2, "bar")],
            ],
            calls,
        )
        del calls[:]
        end = 268435456 - 2
        # see the really-big lookup series in
        # test_lookup_missing_key_before_all_others for details about this
        # assertion.
        self.assertEqual(
            [],
            list(
                bisect_multi_bytes(missing_last_content, 268435456 - 1, ["foo", "bar"])
            ),
        )
        self.assertEqual(
            [
                [(134217727, "foo"), (134217727, "bar")],
                [(201326590, "foo"), (201326590, "bar")],
                [(234881021, "foo"), (234881021, "bar")],
                [(251658236, "foo"), (251658236, "bar")],
                [(260046843, "foo"), (260046843, "bar")],
                [(264241146, "foo"), (264241146, "bar")],
                [(266338297, "foo"), (266338297, "bar")],
                [(267386872, "foo"), (267386872, "bar")],
                [(267911159, "foo"), (267911159, "bar")],
                [(268173302, "foo"), (268173302, "bar")],
                [(268304373, "foo"), (268304373, "bar")],
                [(268369908, "foo"), (268369908, "bar")],
                [(268402675, "foo"), (268402675, "bar")],
                [(268419058, "foo"), (268419058, "bar")],
                [(268427249, "foo"), (268427249, "bar")],
                [(268431344, "foo"), (268431344, "bar")],
                [(268433391, "foo"), (268433391, "bar")],
                [(268434414, "foo"), (268434414, "bar")],
                [(268434925, "foo"), (268434925, "bar")],
                [(268435180, "foo"), (268435180, "bar")],
                [(268435307, "foo"), (268435307, "bar")],
                [(268435370, "foo"), (268435370, "bar")],
                [(268435401, "foo"), (268435401, "bar")],
                [(268435416, "foo"), (268435416, "bar")],
                [(268435423, "foo"), (268435423, "bar")],
                [(268435426, "foo"), (268435426, "bar")],
                [(268435427, "foo"), (268435427, "bar")],
                [(268435428, "foo"), (268435428, "bar")],
                [(268435429, "foo"), (268435429, "bar")],
                [(268435430, "foo"), (268435430, "bar")],
                [(268435431, "foo"), (268435431, "bar")],
                [(268435432, "foo"), (268435432, "bar")],
                [(268435433, "foo"), (268435433, "bar")],
                [(268435434, "foo"), (268435434, "bar")],
                [(268435435, "foo"), (268435435, "bar")],
                [(268435436, "foo"), (268435436, "bar")],
                [(268435437, "foo"), (268435437, "bar")],
                [(268435438, "foo"), (268435438, "bar")],
                [(268435439, "foo"), (268435439, "bar")],
                [(268435440, "foo"), (268435440, "bar")],
                [(268435441, "foo"), (268435441, "bar")],
                [(268435442, "foo"), (268435442, "bar")],
                [(268435443, "foo"), (268435443, "bar")],
                [(268435444, "foo"), (268435444, "bar")],
                [(268435445, "foo"), (268435445, "bar")],
                [(268435446, "foo"), (268435446, "bar")],
                [(268435447, "foo"), (268435447, "bar")],
                [(268435448, "foo"), (268435448, "bar")],
                [(268435449, "foo"), (268435449, "bar")],
                [(268435450, "foo"), (268435450, "bar")],
                [(268435451, "foo"), (268435451, "bar")],
                [(268435452, "foo"), (268435452, "bar")],
                [(268435453, "foo"), (268435453, "bar")],
                [(268435454, "foo"), (268435454, "bar")],
            ],
            calls,
        )

    def test_lookup_when_a_key_is_missing_continues(self):
        calls = []

        def missing_foo_otherwise_missing_first_content(location_keys):
            # returns -1 for all keys unless the byte offset is 0 when it
            # returns False
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[1] == "foo" or location_key[0] == 0:
                    result.append((location_key, False))
                else:
                    result.append((location_key, -1))
            return result

        # given a 2 length file, this should terminate with two calls, one for
        # both keys, and one for bar only.
        self.assertEqual(
            [],
            list(
                bisect_multi_bytes(
                    missing_foo_otherwise_missing_first_content, 2, ["foo", "bar"]
                )
            ),
        )
        self.assertEqual(
            [
                [(1, "foo"), (1, "bar")],
                [(0, "bar")],
            ],
            calls,
        )

    def test_found_keys_returned_other_searches_continue(self):
        calls = []

        def find_bar_at_1_foo_missing_at_0(location_keys):
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key == (1, "bar"):
                    result.append((location_key, "bar-result"))
                elif location_key[0] == 0:
                    result.append((location_key, False))
                else:
                    result.append((location_key, -1))
            return result

        # given a 4 length file, this should terminate with three calls, two for
        # both keys, and one for foo only.
        self.assertEqual(
            [("bar", "bar-result")],
            list(bisect_multi_bytes(find_bar_at_1_foo_missing_at_0, 4, ["foo", "bar"])),
        )
        self.assertEqual(
            [
                [(2, "foo"), (2, "bar")],
                [(1, "foo"), (1, "bar")],
                [(0, "foo")],
            ],
            calls,
        )

    def test_searches_different_keys_in_different_directions(self):
        calls = []

        def missing_bar_at_1_foo_at_3(location_keys):
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[1] == "bar":
                    if location_key[0] == 1:
                        result.append((location_key, False))
                    else:
                        # search down
                        result.append((location_key, -1))
                elif location_key[1] == "foo":
                    if location_key[0] == 3:
                        result.append((location_key, False))
                    else:
                        # search up
                        result.append((location_key, +1))
            return result

        # given a 4 length file, this should terminate with two calls.
        self.assertEqual(
            [], list(bisect_multi_bytes(missing_bar_at_1_foo_at_3, 4, ["foo", "bar"]))
        )
        self.assertEqual(
            [
                [(2, "foo"), (2, "bar")],
                [(3, "foo"), (1, "bar")],
            ],
            calls,
        )

    def test_change_direction_in_single_key_search(self):
        # check that we can search down, up, down again -
        # so length 8, goes 4, 6, 5
        calls = []

        def missing_at_5(location_keys):
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[0] == 5:
                    result.append((location_key, False))
                elif location_key[0] > 5:
                    # search down
                    result.append((location_key, -1))
                else:
                    # search up
                    result.append((location_key, +1))
            return result

        # given a 8 length file, this should terminate with three calls.
        self.assertEqual([], list(bisect_multi_bytes(missing_at_5, 8, ["foo", "bar"])))
        self.assertEqual(
            [
                [(4, "foo"), (4, "bar")],
                [(6, "foo"), (6, "bar")],
                [(5, "foo"), (5, "bar")],
            ],
            calls,
        )



# Copyright (C) 2006, 2008-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the Registry classes."""

import os
import sys

from .. import osutils, registry
from . import TestCase, TestCaseInTempDir


class TestRegistry(TestCase):
    def register_stuff(self, a_registry):
        a_registry.register("one", 1)
        a_registry.register("two", 2)
        a_registry.register("four", 4)
        a_registry.register("five", 5)

    def test_registry(self):
        a_registry = registry.Registry()
        self.register_stuff(a_registry)

        self.assertIsNone(a_registry.default_key)

        # test get() (self.default_key is None)
        self.assertRaises(KeyError, a_registry.get)
        self.assertRaises(KeyError, a_registry.get, None)
        self.assertEqual(2, a_registry.get("two"))
        self.assertRaises(KeyError, a_registry.get, "three")

        # test _set_default_key
        a_registry.default_key = "five"
        self.assertEqual(a_registry.default_key, "five")
        self.assertEqual(5, a_registry.get())
        self.assertEqual(5, a_registry.get(None))
        # If they ask for a specific entry, they should get KeyError
        # not the default value. They can always pass None if they prefer
        self.assertRaises(KeyError, a_registry.get, "six")
        self.assertRaises(KeyError, a_registry._set_default_key, "six")

        # test keys()
        self.assertEqual(["five", "four", "one", "two"], a_registry.keys())

    def test_registry_funcs(self):
        a_registry = registry.Registry()
        self.register_stuff(a_registry)

        self.assertIn("one", a_registry)
        a_registry.remove("one")
        self.assertNotIn("one", a_registry)
        self.assertRaises(KeyError, a_registry.get, "one")

        a_registry.register("one", "one")

        self.assertEqual(["five", "four", "one", "two"], sorted(a_registry.keys()))
        self.assertEqual(
            [("five", 5), ("four", 4), ("one", "one"), ("two", 2)],
            sorted(a_registry.iteritems()),
        )

    def test_register_override(self):
        a_registry = registry.Registry()
        a_registry.register("one", "one")
        self.assertRaises(KeyError, a_registry.register, "one", "two")
        self.assertRaises(
            KeyError, a_registry.register, "one", "two", override_existing=False
        )

        a_registry.register("one", "two", override_existing=True)
        self.assertEqual("two", a_registry.get("one"))

        self.assertRaises(KeyError, a_registry.register_lazy, "one", "three", "four")

        a_registry.register_lazy("one", "module", "member", override_existing=True)

    def test_registry_help(self):
        a_registry = registry.Registry()
        a_registry.register("one", 1, help="help text for one")
        # We should not have to import the module to return the help
        # information
        a_registry.register_lazy(
            "two", "nonexistent_module", "member", help="help text for two"
        )

        # We should be able to handle a callable to get information
        help_calls = []

        def generic_help(reg, key):
            help_calls.append(key)
            return f"generic help for {key}"

        a_registry.register("three", 3, help=generic_help)
        a_registry.register_lazy(
            "four", "nonexistent_module", "member2", help=generic_help
        )
        a_registry.register("five", 5)

        def help_from_object(reg, key):
            obj = reg.get(key)
            return obj.help()

        class SimpleObj:
            def help(self):
                return "this is my help"

        a_registry.register("six", SimpleObj(), help=help_from_object)

        self.assertEqual("help text for one", a_registry.get_help("one"))
        self.assertEqual("help text for two", a_registry.get_help("two"))
        self.assertEqual("generic help for three", a_registry.get_help("three"))
        self.assertEqual(["three"], help_calls)
        self.assertEqual("generic help for four", a_registry.get_help("four"))
        self.assertEqual(["three", "four"], help_calls)
        self.assertEqual(None, a_registry.get_help("five"))
        self.assertEqual("this is my help", a_registry.get_help("six"))

        self.assertRaises(KeyError, a_registry.get_help, None)
        self.assertRaises(KeyError, a_registry.get_help, "seven")

        a_registry.default_key = "one"
        self.assertEqual("help text for one", a_registry.get_help(None))
        self.assertRaises(KeyError, a_registry.get_help, "seven")

        self.assertEqual(
            [
                ("five", None),
                ("four", "generic help for four"),
                ("one", "help text for one"),
                ("six", "this is my help"),
                ("three", "generic help for three"),
                ("two", "help text for two"),
            ],
            sorted((key, a_registry.get_help(key)) for key in a_registry.keys()),
        )

        # We don't know what order it was called in, but we should get
        # 2 more calls to three and four
        self.assertEqual(["four", "four", "three", "three"], sorted(help_calls))

    def test_registry_info(self):
        a_registry = registry.Registry()
        a_registry.register("one", 1, info="string info")
        # We should not have to import the module to return the info
        a_registry.register_lazy("two", "nonexistent_module", "member", info=2)

        # We should be able to handle a callable to get information
        a_registry.register("three", 3, info=["a", "list"])
        obj = object()
        a_registry.register_lazy("four", "nonexistent_module", "member2", info=obj)
        a_registry.register("five", 5)

        self.assertEqual("string info", a_registry.get_info("one"))
        self.assertEqual(2, a_registry.get_info("two"))
        self.assertEqual(["a", "list"], a_registry.get_info("three"))
        self.assertIs(obj, a_registry.get_info("four"))
        self.assertIs(None, a_registry.get_info("five"))

        self.assertRaises(KeyError, a_registry.get_info, None)
        self.assertRaises(KeyError, a_registry.get_info, "six")

        a_registry.default_key = "one"
        self.assertEqual("string info", a_registry.get_info(None))
        self.assertRaises(KeyError, a_registry.get_info, "six")

        self.assertEqual(
            [
                ("five", None),
                ("four", obj),
                ("one", "string info"),
                ("three", ["a", "list"]),
                ("two", 2),
            ],
            sorted((key, a_registry.get_info(key)) for key in a_registry.keys()),
        )

    def test_get_prefix(self):
        my_registry = registry.Registry()
        http_object = object()
        sftp_object = object()
        my_registry.register("http:", http_object)
        my_registry.register("sftp:", sftp_object)
        found_object, suffix = my_registry.get_prefix("http://foo/bar")
        self.assertEqual("//foo/bar", suffix)
        self.assertIs(http_object, found_object)
        self.assertIsNot(sftp_object, found_object)
        found_object, suffix = my_registry.get_prefix("sftp://baz/qux")
        self.assertEqual("//baz/qux", suffix)
        self.assertIs(sftp_object, found_object)

    def test_registry_alias(self):
        a_registry = registry.Registry()
        a_registry.register("one", 1, info="string info")
        a_registry.register_alias("two", "one")
        a_registry.register_alias("three", "one", info="own info")
        self.assertEqual(a_registry.get("one"), a_registry.get("two"))
        self.assertEqual(a_registry.get_help("one"), a_registry.get_help("two"))
        self.assertEqual(a_registry.get_info("one"), a_registry.get_info("two"))
        self.assertEqual("own info", a_registry.get_info("three"))
        self.assertEqual({"two": "one", "three": "one"}, a_registry.aliases())
        self.assertEqual(
            {"one": ["three", "two"]},
            {k: sorted(v) for (k, v) in a_registry.alias_map().items()},
        )

    def test_registry_alias_exists(self):
        a_registry = registry.Registry()
        a_registry.register("one", 1, info="string info")
        a_registry.register("two", 2)
        self.assertRaises(KeyError, a_registry.register_alias, "one", "one")

    def test_registry_alias_targetmissing(self):
        a_registry = registry.Registry()
        self.assertRaises(KeyError, a_registry.register_alias, "one", "two")


class TestRegistryIter(TestCase):
    """Test registry iteration behaviors.

    There are dark corner cases here when the registered objects trigger
    addition in the iterated registry.
    """

    def setUp(self):
        super().setUp()

        # We create a registry with "official" objects and "hidden"
        # objects. The later represent the side effects that led to bug #277048
        # and #430510
        _registry = registry.Registry()

        def register_more():
            _registry.register("hidden", None)

        # Avoid closing over self by binding local variable
        self.registry = _registry
        self.registry.register("passive", None)
        self.registry.register("active", register_more)
        self.registry.register("passive-too", None)

        class InvasiveGetter(registry._ObjectGetter):
            def get_obj(inner_self):  # noqa: N805
                # Surprise ! Getting a registered object (think lazy loaded
                # module) register yet another object !
                _registry.register("more hidden", None)
                return inner_self._obj

        self.registry.register("hacky", None)
        # We peek under the covers because the alternative is to use lazy
        # registration and create a module that can reference our test registry
        # it's too much work for such a corner case -- vila 090916
        self.registry._dict["hacky"] = InvasiveGetter(None)

    def _iter_them(self, iter_func_name):
        iter_func = getattr(self.registry, iter_func_name, None)
        self.assertIsNot(None, iter_func)
        count = 0
        for name, func in iter_func():
            count += 1
            self.assertNotIn(name, ("hidden", "more hidden"))
            if func is not None:
                # Using an object register another one as a side effect
                func()
        self.assertEqual(4, count)

    def test_iteritems(self):
        # the dict is modified during the iteration
        self.assertRaises(RuntimeError, self._iter_them, "iteritems")

    def test_items(self):
        # we should be able to iterate even if one item modify the dict
        self._iter_them("items")


class TestRegistryWithDirs(TestCaseInTempDir):
    """Registry tests that require temporary dirs."""

    def create_plugin_file(self, contents):
        """Create a file to be used as a plugin.

        This is created in a temporary directory, so that we
        are sure that it doesn't start in the plugin path.
        """
        os.mkdir("tmp")
        plugin_name = f"bzr_plugin_a_{osutils.rand_chars(4)}"
        with open("tmp/" + plugin_name + ".py", "wb") as f:
            f.write(contents)
        return plugin_name

    def create_simple_plugin(self):
        return self.create_plugin_file(
            b'object1 = "foo"\n'
            b"\n\n"
            b"def function(a,b,c):\n"
            b"    return a,b,c\n"
            b"\n\n"
            b"class MyClass(object):\n"
            b"    def __init__(self, a):\n"
            b"      self.a = a\n"
            b"\n\n"
        )

    def test_lazy_import_registry_foo(self):
        a_registry = registry.Registry()
        a_registry.register_lazy("foo", "bzrformats.revision", "Revision")
        a_registry.register_lazy("bar", "bzrformats.revision", "NULL_REVISION")
        from bzrformats.revision import NULL_REVISION, Revision

        self.assertEqual(Revision, a_registry.get("foo"))
        self.assertEqual(NULL_REVISION, a_registry.get("bar"))

    def test_lazy_import_registry(self):
        plugin_name = self.create_simple_plugin()
        a_registry = registry.Registry()
        a_registry.register_lazy("obj", plugin_name, "object1")
        a_registry.register_lazy("function", plugin_name, "function")
        a_registry.register_lazy("klass", plugin_name, "MyClass")
        a_registry.register_lazy("module", plugin_name, None)

        self.assertEqual(
            ["function", "klass", "module", "obj"], sorted(a_registry.keys())
        )
        # The plugin should not be loaded until we grab the first object
        self.assertNotIn(plugin_name, sys.modules)

        # By default the plugin won't be in the search path
        self.assertRaises(ImportError, a_registry.get, "obj")

        plugin_path = self.test_dir + "/tmp"  # noqa: S108
        sys.path.append(plugin_path)
        try:
            obj = a_registry.get("obj")
            self.assertEqual("foo", obj)
            self.assertIn(plugin_name, sys.modules)

            # Now grab another object
            func = a_registry.get("function")
            self.assertEqual(plugin_name, func.__module__)
            self.assertEqual("function", func.__name__)
            self.assertEqual((1, [], "3"), func(1, [], "3"))

            # And finally a class
            klass = a_registry.get("klass")
            self.assertEqual(plugin_name, klass.__module__)
            self.assertEqual("MyClass", klass.__name__)

            inst = klass(1)
            self.assertIsInstance(inst, klass)
            self.assertEqual(1, inst.a)

            module = a_registry.get("module")
            self.assertIs(obj, module.object1)
            self.assertIs(func, module.function)
            self.assertIs(klass, module.MyClass)
        finally:
            sys.path.remove(plugin_path)

    def test_lazy_import_get_module(self):
        a_registry = registry.Registry()
        a_registry.register_lazy("obj", "bzrformats.tests.test_registry", "object1")
        self.assertEqual(
            "bzrformats.tests.test_registry", a_registry._get_module("obj")
        )

    def test_normal_get_module(self):
        class AThing:
            """Something."""

        a_registry = registry.Registry()
        a_registry.register("obj", AThing())
        self.assertEqual(
            "bzrformats.tests.test_registry", a_registry._get_module("obj")
        )



# Copyright (C) 2005, 2006, 2007, 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for rio serialization.

A simple, reproducible structured IO format.

rio itself works in Unicode strings.  It is typically encoded to UTF-8,
but this depends on the transport.
"""

import re
from tempfile import TemporaryFile

from .. import rio as _mod_rio
from ..osutils import IterableFile
from ..rio_patch import read_patch_stanza, to_patch_lines
from . import TestCase


def rio_file(stanzas):
    return IterableFile(_mod_rio.rio_iter(stanzas))


class TestRio(TestCase):
    def test_stanza(self):
        """Construct rio stanza in memory."""
        s = _mod_rio.Stanza(number="42", name="fred")
        self.assertIn("number", s)
        self.assertNotIn("color", s)
        self.assertNotIn("42", s)
        self.assertEqual(list(s.iter_pairs()), [("name", "fred"), ("number", "42")])
        self.assertEqual(s.get("number"), "42")
        self.assertEqual(s.get("name"), "fred")

    def test_empty_value(self):
        """Serialize stanza with empty field."""
        s = _mod_rio.Stanza(empty="")
        self.assertEqual(s.to_string(), b"empty: \n")

    def test_to_lines(self):
        """Write simple rio stanza to string."""
        s = _mod_rio.Stanza(number="42", name="fred")
        self.assertEqual(list(s.to_lines()), [b"name: fred\n", b"number: 42\n"])

    def test_as_dict(self):
        """Convert rio Stanza to dictionary."""
        s = _mod_rio.Stanza(number="42", name="fred")
        sd = s.as_dict()
        self.assertEqual(sd, {"number": "42", "name": "fred"})

    def test_to_file(self):
        """Write rio to file."""
        tmpf = TemporaryFile()
        s = _mod_rio.Stanza(
            a_thing='something with "quotes like \\"this\\""', number="42", name="fred"
        )
        s.write(tmpf)
        tmpf.seek(0)
        self.assertEqual(
            tmpf.read(),
            b"""\
a_thing: something with "quotes like \\"this\\""
name: fred
number: 42
""",
        )

    def test_multiline_string(self):
        tmpf = TemporaryFile()
        s = _mod_rio.Stanza(
            motto="war is peace\nfreedom is slavery\nignorance is strength"
        )
        s.write(tmpf)
        tmpf.seek(0)
        self.assertEqual(
            tmpf.read(),
            b"""\
motto: war is peace
\tfreedom is slavery
\tignorance is strength
""",
        )
        tmpf.seek(0)
        s2 = _mod_rio.read_stanza(tmpf)
        self.assertEqual(s, s2)

    def test_read_stanza(self):
        """Load stanza from string."""
        lines = b"""\
revision: mbp@sourcefrog.net-123-abc
timestamp: 1130653962
timezone: 36000
committer: Martin Pool <mbp@test.sourcefrog.net>
""".splitlines(True)
        s = _mod_rio.read_stanza(lines)
        self.assertIn("revision", s)
        self.assertEqual(s.get("revision"), "mbp@sourcefrog.net-123-abc")
        self.assertEqual(
            list(s.iter_pairs()),
            [
                ("revision", "mbp@sourcefrog.net-123-abc"),
                ("timestamp", "1130653962"),
                ("timezone", "36000"),
                ("committer", "Martin Pool <mbp@test.sourcefrog.net>"),
            ],
        )
        self.assertEqual(len(s), 4)

    def test_repeated_field(self):
        """Repeated field in rio."""
        s = _mod_rio.Stanza()
        for k, v in [
            ("a", "10"),
            ("b", "20"),
            ("a", "100"),
            ("b", "200"),
            ("a", "1000"),
            ("b", "2000"),
        ]:
            s.add(k, v)
        s2 = _mod_rio.read_stanza(s.to_lines())
        self.assertEqual(s, s2)
        self.assertEqual(s.get_all("a"), ["10", "100", "1000"])
        self.assertEqual(s.get_all("b"), ["20", "200", "2000"])

    def test_backslash(self):
        s = _mod_rio.Stanza(q="\\")
        t = s.to_string()
        self.assertEqual(t, b"q: \\\n")
        s2 = _mod_rio.read_stanza(s.to_lines())
        self.assertEqual(s, s2)

    def test_blank_line(self):
        s = _mod_rio.Stanza(none="", one="\n", two="\n\n")
        self.assertEqual(
            s.to_string(),
            b"""\
none:\x20
one:\x20
\t
two:\x20
\t
\t
""",
        )
        s2 = _mod_rio.read_stanza(s.to_lines())
        self.assertEqual(s, s2)

    def test_whitespace_value(self):
        s = _mod_rio.Stanza(space=" ", tabs="\t\t\t", combo="\n\t\t\n")
        self.assertEqual(
            s.to_string(),
            b"""\
combo:\x20
\t\t\t
\t
space:\x20\x20
tabs: \t\t\t
""",
        )
        s2 = _mod_rio.read_stanza(s.to_lines())
        self.assertEqual(s, s2)
        self.rio_file_stanzas([s])

    def test_quoted(self):
        """Rio quoted string cases."""
        s = _mod_rio.Stanza(
            q1='"hello"',
            q2=' "for',
            q3='\n\n"for"\n',
            q4='for\n"\nfor',
            q5="\n",
            q6='"',
            q7='""',
            q8="\\",
            q9='\\"\\"',
        )
        s2 = _mod_rio.read_stanza(s.to_lines())
        self.assertEqual(s, s2)
        # apparent bug in read_stanza
        # s3 = _mod_rio.read_stanza(self.stanzas_to_str([s]))
        # self.assertEqual(s, s3)

    def test_read_empty(self):
        """Detect end of rio file."""
        s = _mod_rio.read_stanza([])
        self.assertEqual(s, None)
        self.assertIsNone(s)

    def test_read_nul_byte(self):
        """File consisting of a nul byte causes an error."""
        self.assertRaises(ValueError, _mod_rio.read_stanza, [b"\0"])

    def test_read_nul_bytes(self):
        """File consisting of many nul bytes causes an error."""
        self.assertRaises(ValueError, _mod_rio.read_stanza, [b"\0" * 100])

    def test_read_iter(self):
        """Read several stanzas from file."""
        tmpf = TemporaryFile()
        tmpf.write(
            b"""\
version_header: 1

name: foo
val: 123

name: bar
val: 129319
"""
        )
        tmpf.seek(0)
        reader = _mod_rio.read_stanzas(tmpf)
        stuff = list(reader)
        self.assertEqual(
            stuff,
            [
                _mod_rio.Stanza(version_header="1"),
                _mod_rio.Stanza(name="foo", val="123"),
                _mod_rio.Stanza(name="bar", val="129319"),
            ],
        )

    def test_read_several(self):
        """Read several stanzas from file."""
        tmpf = TemporaryFile()
        tmpf.write(
            b"""\
version_header: 1

name: foo
val: 123

name: quoted
address:   "Willowglen"
\t  42 Wallaby Way
\t  Sydney

name: bar
val: 129319
"""
        )
        tmpf.seek(0)
        s = _mod_rio.read_stanza(tmpf)
        self.assertEqual(s, _mod_rio.Stanza(version_header="1"))
        s = _mod_rio.read_stanza(tmpf)
        self.assertEqual(s, _mod_rio.Stanza(name="foo", val="123"))
        s = _mod_rio.read_stanza(tmpf)
        self.assertEqual(s.get("name"), "quoted")
        self.assertEqual(s.get("address"), '  "Willowglen"\n  42 Wallaby Way\n  Sydney')
        s = _mod_rio.read_stanza(tmpf)
        self.assertEqual(s, _mod_rio.Stanza(name="bar", val="129319"))
        s = _mod_rio.read_stanza(tmpf)
        self.assertEqual(s, None)

    def check_rio_file(self, real_file):
        real_file.seek(0)
        read_write = rio_file(_mod_rio.RioReader(real_file)).read()
        real_file.seek(0)
        self.assertEqual(read_write, real_file.read())

    @staticmethod
    def stanzas_to_str(stanzas):
        return rio_file(stanzas).read()

    def rio_file_stanzas(self, stanzas):
        new_stanzas = list(_mod_rio.RioReader(rio_file(stanzas)))
        self.assertEqual(new_stanzas, stanzas)

    def test_tricky_quoted(self):
        tmpf = TemporaryFile()
        tmpf.write(
            b'''\
s: "one"

s:\x20
\t"one"
\t

s: "

s: ""

s: """

s:\x20
\t

s: \\

s:\x20
\t\\
\t\\\\
\t

s: word\\

s: quote"

s: backslashes\\\\\\

s: both\\\"

'''
        )
        tmpf.seek(0)
        expected_vals = [
            '"one"',
            '\n"one"\n',
            '"',
            '""',
            '"""',
            "\n",
            "\\",
            "\n\\\n\\\\\n",
            "word\\",
            'quote"',
            "backslashes\\\\\\",
            'both\\"',
        ]
        for expected in expected_vals:
            stanza = _mod_rio.read_stanza(tmpf)
            self.rio_file_stanzas([stanza])
            self.assertEqual(len(stanza), 1)
            self.assertEqual(stanza.get("s"), expected)

    def test_write_empty_stanza(self):
        """Write empty stanza."""
        l = list(_mod_rio.Stanza().to_lines())
        self.assertEqual(l, [])

    def test_rio_raises_type_error(self):
        """TypeError on adding invalid type to Stanza."""
        s = _mod_rio.Stanza()
        self.assertRaises(TypeError, s.add, "foo", {})

    def test_rio_raises_type_error_key(self):
        """TypeError on adding invalid type to Stanza."""
        s = _mod_rio.Stanza()
        self.assertRaises(TypeError, s.add, 10, {})

    def test_rio_surrogateescape(self):
        raw_bytes = b"\xcb"
        self.assertRaises(UnicodeDecodeError, raw_bytes.decode, "utf-8")
        try:
            uni_data = raw_bytes.decode("utf-8", "surrogateescape")
        except LookupError:
            self.skipTest("surrogateescape is not available on Python < 3")
        try:
            _mod_rio.Stanza(foo=uni_data)
        except TypeError:
            pass
        else:
            self.fail()

    def test_rio_unicode(self):
        uni_data = "\N{KATAKANA LETTER O}"
        s = _mod_rio.Stanza(foo=uni_data)
        self.assertEqual(s.get("foo"), uni_data)
        raw_lines = s.to_lines()
        self.assertEqual(raw_lines, [b"foo: " + uni_data.encode("utf-8") + b"\n"])
        new_s = _mod_rio.read_stanza(raw_lines)
        self.assertEqual(new_s.get("foo"), uni_data)

    def mail_munge(self, lines, dos_nl=True):
        new_lines = []
        for line in lines:
            line = re.sub(b" *\n", b"\n", line)
            if dos_nl:
                line = re.sub(b"([^\r])\n", b"\\1\r\n", line)
            new_lines.append(line)
        return new_lines

    def test_patch_rio(self):
        stanza = _mod_rio.Stanza(data="#\n\r\\r ", space=" " * 255, hash="#" * 255)
        lines = to_patch_lines(stanza)
        for line in lines:
            self.assertContainsRe(line, b"^# ")
            self.assertGreaterEqual(72, len(line))
        for line in to_patch_lines(stanza, max_width=12):
            self.assertGreaterEqual(12, len(line))
        new_stanza = read_patch_stanza(self.mail_munge(lines, dos_nl=False))
        lines = self.mail_munge(lines)
        new_stanza = read_patch_stanza(lines)
        self.assertEqual("#\n\r\\r ", new_stanza.get("data"))
        self.assertEqual(" " * 255, new_stanza.get("space"))
        self.assertEqual("#" * 255, new_stanza.get("hash"))

    def test_patch_rio_linebreaks(self):
        stanza = _mod_rio.Stanza(breaktest="linebreak -/" * 30)
        line1 = to_patch_lines(stanza, 71)[0]
        self.assertContainsRe(line1, b"linebreak\\\\\n")
        stanza = _mod_rio.Stanza(breaktest="linebreak-/" * 30)
        self.assertContainsRe(to_patch_lines(stanza, 70)[0], b"linebreak-\\\\\n")
        stanza = _mod_rio.Stanza(breaktest="linebreak/" * 30)
        self.assertContainsRe(to_patch_lines(stanza, 70)[0], b"linebreak\\\\\n")


class TestValidTag(TestCase):
    def test_ok(self):
        self.assertTrue(_mod_rio.valid_tag("foo"))

    def test_no_spaces(self):
        self.assertFalse(_mod_rio.valid_tag("foo bla"))

    def test_numeric(self):
        self.assertTrue(_mod_rio.valid_tag("3foo423"))

    def test_no_colon(self):
        self.assertFalse(_mod_rio.valid_tag("foo:bla"))

    def test_type_error(self):
        self.assertRaises(TypeError, _mod_rio.valid_tag, 423)

    def test_empty(self):
        self.assertFalse(_mod_rio.valid_tag(""))

    def test_unicode(self):
        # When str is a unicode type, it is valid for a tag
        self.assertTrue(_mod_rio.valid_tag("foo"))

    def test_non_ascii_char(self):
        self.assertFalse(_mod_rio.valid_tag("\xb5"))


class TestReadUTF8Stanza(TestCase):
    def assertReadStanza(self, result, line_iter):
        s = _mod_rio.read_stanza(line_iter)
        self.assertEqual(result, s)
        if s is not None:
            for tag, value in s.iter_pairs():
                self.assertIsInstance(tag, str)
                self.assertIsInstance(value, str)

    def assertReadStanzaRaises(self, exception, line_iter):
        self.assertRaises(exception, _mod_rio.read_stanza, line_iter)

    def test_no_string(self):
        self.assertReadStanzaRaises(TypeError, [21323])

    def test_empty(self):
        self.assertReadStanza(None, [])

    def test_none(self):
        self.assertReadStanza(None, [b""])

    def test_simple(self):
        self.assertReadStanza(_mod_rio.Stanza(foo="bar"), [b"foo: bar\n", b""])

    def test_multi_line(self):
        self.assertReadStanza(
            _mod_rio.Stanza(foo="bar\nbla"), [b"foo: bar\n", b"\tbla\n"]
        )

    def test_repeated(self):
        s = _mod_rio.Stanza()
        s.add("foo", "bar")
        s.add("foo", "foo")
        self.assertReadStanza(s, [b"foo: bar\n", b"foo: foo\n"])

    def test_invalid_early_colon(self):
        self.assertReadStanzaRaises(ValueError, [b"f:oo: bar\n"])

    def test_invalid_tag(self):
        self.assertReadStanzaRaises(ValueError, [b"f%oo: bar\n"])

    def test_continuation_too_early(self):
        self.assertReadStanzaRaises(ValueError, [b"\tbar\n"])

    def test_large(self):
        value = b"bla" * 9000
        self.assertReadStanza(
            _mod_rio.Stanza(foo=value.decode()), [b"foo: %s\n" % value]
        )

    def test_non_ascii_char(self):
        self.assertReadStanza(
            _mod_rio.Stanza(foo="n\xe5me"), ["foo: n\xe5me\n".encode()]
        )



# Copyright (C) 2005, 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for different inventory implementations."""

# NOTE: Don't import Inventory here, to make sure that we don't accidentally
# hardcode that when we should be using self.make_inventory

from bzrformats import inventory, osutils
from bzrformats.errors import InconsistentDelta
from bzrformats.inventory import NoSuchId
from bzrformats.tests.per_inventory import TestCaseWithInventory

from ...inventory import InventoryFile, InventoryLink
from ...inventory_delta import InventoryDelta


class TestInventory(TestCaseWithInventory):
    def make_init_inventory(self):
        inv = inventory.Inventory(root_id=None, revision_id=b"initial-rev")
        root = inventory.InventoryDirectory(b"tree-root", "", None, b"initial-rev")
        inv.add(root)
        return self.inv_to_test_inv(inv)

    def make_file(
        self, file_id, name, parent_id, content=b"content\n", revision=b"new-test-rev"
    ):
        return InventoryFile(
            file_id,
            name,
            parent_id,
            text_sha1=osutils.sha_string(content),
            text_size=len(content),
            revision=revision,
        )

    def make_link(self, file_id, name, parent_id, target="link-target\n"):
        return InventoryLink(file_id, name, parent_id, symlink_target=target)

    def prepare_inv_with_nested_dirs(self):
        inv = inventory.Inventory(root_id=None)
        root = inventory.InventoryDirectory(b"tree-root", "", None, b"revision")
        inv.add(root)
        for args in [
            ("src", "directory", b"src-id"),
            ("doc", "directory", b"doc-id"),
            ("src/hello.c", "file", b"hello-id"),
            ("src/bye.c", "file", b"bye-id"),
            ("zz", "file", b"zz-id"),
            ("src/sub/", "directory", b"sub-id"),
            ("src/zz.c", "file", b"zzc-id"),
            ("src/sub/a", "file", b"a-id"),
            ("Makefile", "file", b"makefile-id"),
        ]:
            kwargs = {}
            if args[1] == "file":
                kwargs["text_sha1"] = osutils.sha_string(b"content\n")
                kwargs["text_size"] = len(b"content\n")
            inv.add_path(*args, revision=b"revision", **kwargs)
        return self.inv_to_test_inv(inv)


class TestInventoryCreateByApplyDelta(TestInventory):
    """A subset of the inventory delta application tests.

    See test_inv which has comprehensive delta application tests for
    inventories, dirstate, and repository based inventories.
    """

    def test_add(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta(
            InventoryDelta(
                [
                    (None, "a", b"a-id", self.make_file(b"a-id", "a", b"tree-root")),
                ]
            ),
            b"new-test-rev",
        )
        self.assertEqual("a", inv.id2path(b"a-id"))

    def test_delete(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta(
            InventoryDelta(
                [
                    (None, "a", b"a-id", self.make_file(b"a-id", "a", b"tree-root")),
                ]
            ),
            b"new-rev-1",
        )
        self.assertEqual("a", inv.id2path(b"a-id"))
        inv = inv.create_by_apply_delta(
            InventoryDelta(
                [
                    ("a", None, b"a-id", None),
                ]
            ),
            b"new-rev-2",
        )
        self.assertRaises(NoSuchId, inv.id2path, b"a-id")

    def test_rename(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta(
            InventoryDelta(
                [
                    (None, "a", b"a-id", self.make_file(b"a-id", "a", b"tree-root")),
                ]
            ),
            b"new-rev-1",
        )
        self.assertEqual("a", inv.id2path(b"a-id"))
        a_ie = inv.get_entry(b"a-id")
        b_ie = self.make_file(a_ie.file_id, "b", a_ie.parent_id)
        inv = inv.create_by_apply_delta(
            InventoryDelta([("a", "b", b"a-id", b_ie)]), b"new-rev-2"
        )
        self.assertEqual("b", inv.id2path(b"a-id"))

    def test_illegal(self):
        # A file-id cannot appear in a delta more than once
        inv = self.make_init_inventory()
        self.assertRaises(
            InconsistentDelta,
            inv.create_by_apply_delta,
            InventoryDelta(
                [
                    (None, "a", b"id-1", self.make_file(b"id-1", "a", b"tree-root")),
                    (None, "b", b"id-1", self.make_file(b"id-1", "b", b"tree-root")),
                ]
            ),
            b"new-rev-1",
        )


class TestInventoryReads(TestInventory):
    def test_is_root(self):
        """Ensure our root-checking code is accurate."""
        inv = self.make_init_inventory()
        self.assertTrue(inv.is_root(b"tree-root"))
        self.assertFalse(inv.is_root(b"booga"))
        ie = inventory.InventoryDirectory(
            b"booga", "", None, revision=inv.root.revision
        )
        inv = inv.create_by_apply_delta(
            InventoryDelta([("", None, b"tree-root", None), (None, "", b"booga", ie)]),
            b"new-rev-2",
        )
        self.assertFalse(inv.is_root(b"TREE_ROOT"))
        self.assertTrue(inv.is_root(b"booga"))

    def test_ids(self):
        """Test detection of files within selected directories."""
        inv = inventory.Inventory(root_id=None)
        root = inventory.InventoryDirectory(b"tree-root", "", None, b"revision")
        inv.add(root)
        for args in [
            ("src", "directory", b"src-id"),
            ("doc", "directory", b"doc-id"),
            ("src/hello.c", "file"),
            ("src/bye.c", "file", b"bye-id"),
            ("Makefile", "file"),
        ]:
            kwargs = {}
            if args[1] == "file":
                kwargs["text_sha1"] = osutils.sha_string(b"content\n")
                kwargs["text_size"] = len(b"content\n")
            inv.add_path(*args, revision=b"revision", **kwargs)
        inv = self.inv_to_test_inv(inv)
        self.assertEqual(inv.path2id("src"), b"src-id")
        self.assertEqual(inv.path2id("src/bye.c"), b"bye-id")

    def test_get_entry_by_path_partial(self):
        inv = inventory.Inventory(root_id=None)
        root = inventory.InventoryDirectory(b"TREE_ROOT", "", None, b"revision")
        inv.add(root)
        for args in [
            ("src", "directory", b"src-id"),
            ("doc", "directory", b"doc-id"),
            ("src/hello.c", "file"),
            ("src/bye.c", "file", b"bye-id"),
            ("Makefile", "file"),
            ("external", "tree-reference", b"other-root"),
        ]:
            kwargs = {}
            if args[1] == "file":
                kwargs["text_sha1"] = osutils.sha_string(b"content\n")
                kwargs["text_size"] = len(b"content\n")
            if args[1] == "tree-reference":
                kwargs["reference_revision"] = b"reference"
            ie = inv.add_path(*args, revision=b"revision", **kwargs)
        inv = self.inv_to_test_inv(inv)

        # Standard lookups
        ie, resolved, remaining = inv.get_entry_by_path_partial("")
        self.assertEqual((ie.file_id, resolved, remaining), (b"TREE_ROOT", [], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial("src")
        self.assertEqual((ie.file_id, resolved, remaining), (b"src-id", ["src"], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial("src/bye.c")
        self.assertEqual(
            (ie.file_id, resolved, remaining), (b"bye-id", ["src", "bye.c"], [])
        )

        # Paths in the external tree
        ie, resolved, remaining = inv.get_entry_by_path_partial("external")
        self.assertEqual(
            (ie.file_id, resolved, remaining), (b"other-root", ["external"], [])
        )
        ie, resolved, remaining = inv.get_entry_by_path_partial("external/blah")
        self.assertEqual(
            (ie.file_id, resolved, remaining), (b"other-root", ["external"], ["blah"])
        )

        # Nonexistant paths
        ie, resolved, remaining = inv.get_entry_by_path_partial("foo.c")
        self.assertEqual((ie, resolved, remaining), (None, None, None))

    def test_non_directory_children(self):
        """Test path2id when a parent directory has no children."""
        inv = inventory.Inventory(b"tree-root")
        inv.add(self.make_file(b"file-id", "file", b"tree-root"))
        inv.add(self.make_link(b"link-id", "link", b"tree-root"))
        self.assertIs(None, inv.path2id("file/subfile"))
        self.assertIs(None, inv.path2id("link/subfile"))

    def test_is_unmodified(self):
        f1 = self.make_file(b"file-id", "file", b"tree-root", revision=b"rev")
        self.assertTrue(f1.is_unmodified(f1))
        f2 = self.make_file(b"file-id", "file", b"tree-root", revision=b"rev")
        self.assertTrue(f1.is_unmodified(f2))
        f3 = self.make_file(b"file-id", "file", b"tree-root")
        self.assertFalse(f1.is_unmodified(f3))
        f4 = self.make_file(b"file-id", "file", b"tree-root", revision=b"rev1")
        self.assertFalse(f1.is_unmodified(f4))

    def test_iter_entries(self):
        inv = self.prepare_inv_with_nested_dirs()

        # Test all entries
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
                ("src/zz.c", b"zzc-id"),
                ("zz", b"zz-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries()],
        )

        # Test a subdirectory
        self.assertEqual(
            [
                ("bye.c", b"bye-id"),
                ("hello.c", b"hello-id"),
                ("sub", b"sub-id"),
                ("sub/a", b"a-id"),
                ("zz.c", b"zzc-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries(from_dir=b"src-id")],
        )

        # Test not recursing at the root level
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("zz", b"zz-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries(recursive=False)],
        )

        # Test not recursing at a subdirectory level
        self.assertEqual(
            [
                ("bye.c", b"bye-id"),
                ("hello.c", b"hello-id"),
                ("sub", b"sub-id"),
                ("zz.c", b"zzc-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries(from_dir=b"src-id", recursive=False)
            ],
        )

    def test_iter_entries_by_dir(self):
        inv = self.prepare_inv_with_nested_dirs()
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("zz", b"zz-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/zz.c", b"zzc-id"),
                ("src/sub/a", b"a-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir()],
        )
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("zz", b"zz-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/zz.c", b"zzc-id"),
                ("src/sub/a", b"a-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids={
                        b"a-id",
                        b"zzc-id",
                        b"doc-id",
                        b"tree-root",
                        b"hello-id",
                        b"bye-id",
                        b"zz-id",
                        b"src-id",
                        b"makefile-id",
                        b"sub-id",
                    }
                )
            ],
        )

        self.assertEqual(
            [
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("zz", b"zz-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/zz.c", b"zzc-id"),
                ("src/sub/a", b"a-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids={
                        b"a-id",
                        b"zzc-id",
                        b"doc-id",
                        b"hello-id",
                        b"bye-id",
                        b"zz-id",
                        b"makefile-id",
                    }
                )
            ],
        )

        self.assertEqual(
            [
                ("Makefile", b"makefile-id"),
                ("src/bye.c", b"bye-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids={b"bye-id", b"makefile-id"}
                )
            ],
        )

        self.assertEqual(
            [
                ("Makefile", b"makefile-id"),
                ("src/bye.c", b"bye-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids={b"bye-id", b"makefile-id"}
                )
            ],
        )

        self.assertEqual(
            [
                ("src/bye.c", b"bye-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(specific_file_ids={b"bye-id"})
            ],
        )


class TestInventoryFiltering(TestInventory):
    def test_inv_filter_empty(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter(set())
        self.assertEqual(
            [
                ("", b"tree-root"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_files(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter({b"zz-id", b"hello-id", b"a-id"})
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("src", b"src-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
                ("zz", b"zz-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_dirs(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter({b"doc-id", b"sub-id"})
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_files_and_dirs(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter({b"makefile-id", b"src-id"})
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("src", b"src-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
                ("src/zz.c", b"zzc-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_entry_not_present(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter({b"not-present-id"})
        self.assertEqual(
            [
                ("", b"tree-root"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )
